/* 
 * the BSD socket syscalls
 *
 * Consists of 4.4BSD socket API that either take file
 * descriptors as an argument (send()/recv() etc) or return new
 * file descriptors (accept(), socket()). Note that poll() and select()
 * are (strictly speaking) not members of the BSD socket API and
 * are located elsewhere.  
*/
#include <sys/socket.h>

#include "target.h"   /* FAR */
#include "sofdmux.h"  /* __sock_fd_*() */
#include "bsderrno.h" /* bsd error numbers */
#include "fsext.h"    /* __fsext_setup() */
#include "syscall.h"  /* sas_*() */

int socket(int domain, int stype, int proto)
{
  if (__fsext_setup() != -1) /* sets errno on failure */
  {
    if (__sock_fd_reserve() != -1) /* sets errno on failure */
    {
      struct fdmux_object FAR *socket = bsdsys_socket( domain, stype, proto );
      if (socket) /* otherwise errno will already be set */
      {
        int fd = __sock_fd_alloc(socket);
        if (fd != -1)
          return fd;
        fd = errno;
        bsdsys_close(socket);
        errno = fd;
      }
    }
  }
  return -1;
}

int accept(int fd, struct sockaddr *addr, int *addrlen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (!socket)
    errno = ENOTSOCK;
  else if (__sock_fd_reserve() != -1) /* sets errno on failure */
  {
    socket = bsdsys_accept(socket, addr, addrlen);
    if (socket)
    {
      int fd = __sock_fd_alloc(socket);
      if (fd != -1)
        return fd;
      fd = errno;
      bsdsys_close(socket);
      errno = fd;
    }
  }
  return -1;  
}


int bind(int fd, const struct sockaddr *addr, int namelen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_bind(socket, addr, namelen);
  errno = ENOTSOCK;
  return -1;
}    
  

int connect(int fd, const struct sockaddr *name, int namelen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_connect(socket, name, namelen);
  errno = ENOTSOCK;
  return -1;
}


int getpeername(int fd, struct sockaddr *name, int *namelen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_getpeername(socket, name, namelen);
  errno = ENOTSOCK;
  return -1;
}


int getsockname(int fd, struct sockaddr *name, int * namelen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_getsockname(socket, name, namelen);
  errno = ENOTSOCK;
  return -1;
}  


int listen(int fd, int backlog)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_listen(socket, backlog);
  errno = ENOTSOCK;
  return -1;
}  

int recvmsg (int fd, struct msghdr *msg, int flags )
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_recvmsg (socket, msg, flags);
  errno = ENOTSOCK;
  return -1;
}

int recvfrom(int fd, char *buf, int len, int flags,
             struct sockaddr *from, int *fromlen)
{
  int rc;
  struct msghdr msg;
  struct iovec iov;
  iov.iov_base = buf;
  iov.iov_len = len;
  memset( &msg, 0, sizeof(msg));
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  if (fromlen)
  {
    VERIFY_RW(fromlen,sizeof(*fromlen));
    msg.msg_name = (char *)from;
    msg.msg_namelen = *fromlen;
  }
  rc = recvmsg (fd, &msg, flags);
  if (rc != -1 && msg.msg_name)
    *fromlen = msg.msg_namelen;
  return rc;
}   

int recv(int fd, char *buf, int len, int flags)
{
  return recvfrom(fd, buf, len, flags, 0, 0 );
}

int sendmsg (int fd, const struct msghdr *msg, int flags )
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_sendmsg (socket, msg, flags);
  errno = ENOTSOCK;
  return -1;
}

int sendto (int fd, const char *buf, int len, int flags,
             const struct sockaddr *to, int tolen)
{
  struct msghdr msg;
  struct iovec iov;
  iov.iov_base = (char *)(*((void **)&buf));
  iov.iov_len = len;
  memset( &msg, 0, sizeof(msg));
  msg.msg_name = (char *)(*((void **)&to));
  msg.msg_namelen = tolen;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  return sendmsg (fd, &msg, flags);
}  

int send (int fd, const char *buf, int len, int flags)
{
  return sendto ( fd, buf, len, flags, 0, 0 );
}


int setsockopt (int fd, int level, int optname,
                const char *optval, int optlen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_setsockopt (socket, level, optname, optval, optlen);
  errno = ENOTSOCK;
  return -1;
}  


int getsockopt (int fd, int level, int optname,
               char *optval, int *optlen)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_getsockopt (socket, level, optname, optval, optlen);
  errno = ENOTSOCK;
  return -1;
}  


int shutdown (int fd, int how)
{
  struct fdmux_object FAR *socket = __sock_fd_lookup(fd);
  if (socket)
    return bsdsys_shutdown (socket, how);
  errno = ENOTSOCK;
  return -1;
}

