/*
** getservbyname() and getservbyport()
** 
** Created June 2000 by Cyrus Patel <cyp@fb14.uni-mainz.de>
*/
#include <netdb.h>   /* h_errno */

struct my_servent {
    struct servent se;
    char namebuf[32];
    char aliasbuf[64];
    char *aliaslist[5];
    int protonum;
};

#define NDBRGEN_PAYLOAD        struct my_servent  /* what we deal with internally*/
#define NDBRGEN_RETTYPE        struct servent  /* what we return (not ptr!) */
#define NDBRGEN_ETCFNCONF      "SERVICES"      /* entry in wattcp.cfg */
#define NDBRGEN_ENDENTFXNN     endservent     /* what to generate for endXXXent() */
#define NDBRGEN_GETENTFXNN     getservent     /* what to generate for getXXXent() */
#define NDBRGEN_SETENTFXNN     setservent     /* what to generate for setXXXent() */
#define NDBRGEN_SETCANSTAYOPEN 1  /* does setXXXent take a 'stayopen' arg? */
#define NDBRGEN_HAVEALIASES    1  /* yes, we need alias support */

#include "dbentgen.h" /* generate all the necessary db access skeleton code */ 
#include <netinet/in.h>  /* IPPROTO_xxx */

/* ======================================================================= */

/* What protos are supported is determined here */
static const char *check_proto( const char *name, int *protoP)
{
  static struct { const char *name; short num; } prototab[] = {
                  { "udp", IPPROTO_UDP },
                  { "tcp", IPPROTO_TCP } };
  unsigned int i;
  for (i = 0; i < (sizeof(prototab)/sizeof(prototab[0])); i++)
  {
    if (strcmp( name, prototab[i].name )==0)
    {
      *protoP = prototab[i].num;
      return prototab[i].name;
    }
  }
  return (const char *)0;
}

/* ------------------------------------------------------------------- */

struct servent * getservbyname (const char * name, const char *proto)
{
  int scan_proto = -1;

  if (name)
  {
    scan_proto = 0;
    if (proto) /* if no proto is specified, all protos match */
    {
      if (!check_proto(proto, &scan_proto))
        scan_proto = -1;
    }
  }

  if (scan_proto >= 0 && scan_proto <= 255)
  {    
    /* begin enumeration by passing NULL to enum_db_entries */
    struct my_servent *e = enum_db_entries((struct my_servent *)0);
    while (e)
    {
      if (scan_proto == 0 || (scan_proto == e->protonum))
      {
        unsigned int i;
        if (strcmp(name, e->namebuf)==0)
          return &(e->se);
        i = 0;
        while (e->aliaslist[i])
        {
          if (strcmp(name, e->aliaslist[i])==0)
            return &(e->se);
          i++;
        }
      }
      e = enum_db_entries(e); /* get next */
    }
  }

  h_errno = NO_DATA;
  return (struct servent *)0;
}

/* ---------------------------------------------------------------------- */

struct servent * getservbyport (int port, const char *proto)
{
  int scan_proto = -1;

  if (port > 0 && ((unsigned int)port) < 0xffffu)
  {
    scan_proto = 0;
    if (proto) /* if no proto is specified, all protos match */
    {
      if (!check_proto(proto, &scan_proto))
        scan_proto = -1;
    }
  }

  if (scan_proto >= 0 && scan_proto <= 255)
  {
    /* begin enumeration by passing NULL to enum_db_entries */
    struct my_servent *e = enum_db_entries((struct my_servent *)0);
    while (e)
    {
      if (scan_proto == 0 || (scan_proto == e->protonum))
      {
        if (port == e->se.s_port)
          return &(e->se);
      }
      e = enum_db_entries(e); /* get next */
    }
  }

  h_errno = NO_DATA;
  return (struct servent *)0;
}

/* ====================================================================== */

static const char *__parse_port(const char *str, short *portP, int *protoP)
{
  unsigned int port = 0;
  while (*str >= '0' && *str <= '9' && port < 6553)
  {
    port = (port * 10) + (*str++ - '0');
  }
  if (port > 0 && *str++ == '/')
  {
    const char *protoname = check_proto(str, protoP);
    if (protoname)
    {
      *portP = htons((short)port);
      return protoname;
    }
  }
  return (const char *)0;
}  

/* ---------------------------------------------------------------------- */

/* This is our entry parser called by the auto-generated code to 
** parse a line in "/etc/services". 
*/
static NDBRGEN_PAYLOAD *parser_cb(const char *tokens[], int numtoks, 
                                  NDBRGEN_PAYLOAD *resP)
{
  NDBRGEN_PAYLOAD *e = (NDBRGEN_PAYLOAD *)0;
  if (numtoks >= 2)
  {
    if (strlen(tokens[0]) < (sizeof(e->namebuf)-1))
    {
      int proto; short port; 
      const char *protoname = __parse_port(tokens[1], &port, &proto);
      if (protoname)
      {
        e = resP;
        strcpy( e->namebuf, tokens[0] );
        e->protonum    = proto;
        e->aliaslist[0]= (char *)0;

        e->se.s_name   = e->namebuf;
        e->se.s_aliases= &(e->aliaslist[0]);
        e->se.s_port   = port;
        e->se.s_proto  = (char *)protoname; 

        db_add_aliases( &tokens[2], numtoks-2, e->aliasbuf, sizeof(e->aliasbuf),
             &(e->aliaslist[0]), sizeof(e->aliaslist)/sizeof(e->aliaslist[0]));
      }
    }
  }
  return e;
}


#if defined(TEST_PROG)
static void dump_ent(const char *label, const struct servent *e)
{
  if (!e)
    printf("%s: no data!\n", label);
  else
  {
    int i;
    printf ("%s: s_name='%s' s_port='%d' s_proto='%s'\n",
           label, e->s_name, htons(e->s_port), e->s_proto );
    for (i = 0; e->s_aliases && e->s_aliases[i]; i++)
       printf ("        s_aliases[%d]='%s'\n", i, e->s_aliases[i]);
  }
  return;
}

int main (int argc, char *argv[])
{
  if (argc > 1)
  {
    if (strcmp(argv[1],"all")==0)
      argc = 1; /* fall through */
    else 
    {
      char *proto = argv[1];
      while (*proto && *proto!='/')
        proto++;
      if (*proto != '/')
        proto = NULL;
      else
        *proto++ = '\0';

      if (*argv[1] >= '0' && *argv[1] <= '9')
        dump_ent("byport", getservbyport( htons(atoi(argv[1])), proto ));
      else
        dump_ent("byname", getservbyname( argv[1], proto ));
    }
  }
  if (argc == 1)
  {
    struct servent *e;
    while ((e = getservent()) != NULL)
      dump_ent("getent", e);
  }
  return (0);
}
#endif /* TEST_PROG */


